"""TODO import module that are needed:
Make sure not to run it as pytest suite"""

import numba
import numpy as np


@numba.jit("f8[:, :, :](f8[:, :], f8[:, :], f8[:, :], f8[:])")
def _rbf_laplacian_debug(K, S, S_in, S_out):
    p, n = K.shape
    out = np.zeros((p, p, n), dtype=np.float64)
    for i in range(p):
        for j in range(p):
            for k in range(n):
                tmp = S_out[k] - S[i, k] - S[j, k] + S_in[i, j]
                out[i, j, k] += K[i, k] * K[j, k] * tmp
    return out


@numba.jit("f8[:, :, :](f8[:, :], f8[:, :], f8[:, :], f8[:])")
def _exp_laplacian_debug(K, S, S_in, S_out):
    p, n = K.shape
    out = np.zeros((p, p, n), dtype=np.float64)
    for i in range(p):
        for j in range(p):
            for k in range(n):
                norm = (S_out[k] - 2*S[i, k] + S_in[i, i])
                norm *= (S_out[k] - 2*S[j, k] + S_in[j, j])
                tmp = S_out[k] - S[i, k] - S[j, k] + S_in[i, j]
                if norm < 0:
                    norm = 1
                    tmp = 0
                out[i, j, k] += K[i, k] * K[j, k] * tmp / np.sqrt(norm)
    return out

class myFunc:
    """Generate random function and check for derivatives calculation"""
    def __init__(self, d=1, sigma=1, n=10, kernel='gaussian'):
        self.x = np.random.randn(n, d)
        self.alpha = np.random.randn(n)
        self.sigma = sigma
        if kernel == 'gaussian':
            self.name = 'gaussian'
            self.kernel = rbf_kernel
            self.laplacian = _rbf_laplacian_debug
            self.norm = 2 / sigma ** 2
        else:
            self.name = 'exponential'
            self.kernel = exp_kernel
            self.laplacian = _exp_laplacian_debug
            self.norm = 1 / sigma


    @staticmethod
    def format(x):
        if type(x)!= np.ndarray:
            x = np.array(x)
        if len(x.shape) == 0:
            x = x[np.newaxis, np.newaxis]
        if len(x.shape) == 1:
            x = x[:, np.newaxis]
        return x

    def __call__(self, x):
        x = self.format(x) 
        k_x = self.kernel(x, self.x, sigma=self.sigma)
        out = k_x @ self.alpha
        return out.squeeze()

    def deriv(self, x):
        x = self.format(x)
        out = np.zeros((len(x), self.x.shape[1]))
        k = self.kernel(x, self.x, sigma=self.sigma)
        for i in range(self.x.shape[0]):
            tmp = self.x[i] - x
            if self.name == 'exponential':
                norm = np.sqrt(np.sum(tmp ** 2, axis=1)[:, np.newaxis])
                tmp /= norm
            out += tmp * k[:, i:i+1] * self.alpha[i]
        out *= self.norm
        return out.squeeze()
        
    def norm_gradient(self, x):
        out = self.deriv(x) 
        out **= 2
        if len(out.shape) > 1:
            out = np.sum(out, axis=1)
        return out

    def norm_gradient_laplacian(self, x):
        x = self.format(x)
        K = self.kernel(self.x, x, sigma=self.sigma)
        S_in = scalar_product(self.x, self.x)
        S = scalar_product(self.x, x)
        S_out = np.sum(x ** 2, axis=1)
        L = self.laplacian(K, S, (S_in + S_in.T) / 2, S_out)
        L *= self.norm ** 2
        L = (L + L.transpose(1, 0, 2)) / 2
        out = np.zeros(len(x))
        for i in range(len(self.alpha)):
            for j in range(len(self.alpha)):
                out += self.alpha[i] * self.alpha[j] * L[i, j]
        return out


def test_derivative():
    np.random.seed(0)
    d = 50
    f = myFunc(sigma=1, d=d, kernel='exponential')
    x = np.random.randn(1000, d)
    a = f.norm_gradient(x)
    b = f.norm_gradient_laplacian(x)
    assert (np.abs(a - b) / np.abs(a + b)).max() < 1e-7



"""
Fast implementation of operator at play to estimate Laplacian eigendecomposition.

@ Vivien Cabannes, 2023

Notes
-----
Could be nice to implement it in pytorch to run it on GPUs.
"""
import numba
import numpy as np
from .helper import (
    distance,
    distance_square,
    scalar_product,
)


def rbf_kernel(x1, x2, sigma: float = 1):
    r"""
    Computation of the Gaussian kernel
    .. math::
        K[i,j] = \exp(-\| x1[i, :] - x2[j,:] / \sigma \|^2)

    Parameters
    ----------
    x1: ndarray of size (n, d)
    x2: ndarray of size (n, d)
    sigma
        Bandwidth parameter for the kernel

    Returns
    -------
    K: ndarray of size (n, n)
    """
    K = distance_square(x1, x2)
    K /= -(sigma**2)
    np.exp(K, out=K)
    return K


def exp_kernel(x1, x2, sigma: float = 1):
    r"""
    Computation of the exponential kernel
    .. math::
        K[i,j] = \exp(-\| x1[i, :] - x2[j,:] / \sigma \|)

    Parameters
    ----------
    x1: ndarray of size (n, d)
    x2: ndarray of size (n, d)
    sigma
        Bandwidth parameter for the kernel

    Returns
    -------
    K: ndarray of size (n, n)
    """
    K = distance(x1, x2)
    K /= -sigma
    np.exp(K, out=K)
    return K


def rbf_laplacian(x_repr, x, sigma: float = 1, K=None):
    r"""
    Computation of the discrete Laplacian operator for the gaussian kernel
    .. math::
        L[i,j] = \sum_{k} \nabla_{x} k(x_k, y_i)^\top \nabla_x k(x_i, y_j)

    Parameters
    ----------
    x_repr: ndarray of size (p, d)
        ::math:`x_repr = y` representer to discretize `L^2`
    x: ndarray of size (n, d)
        Data to estimate the distribution on `L^2(X)`
    sigma
        Bandwidth parameter for the kernel
    K: ndarray of size (p, n) (optional, default is None)
        Pre-computation of the gaussian kernel `k(x_repr, x)`

    Returns
    -------
    L: ndarray of size (p, p)
    """
    if K is None:
        K = rbf_kernel(x_repr, x, sigma=sigma)
    S_in = scalar_product(x_repr, x_repr)
    S = scalar_product(x_repr, x)
    S_out = np.sum(x**2, axis=1)
    L = _rbf_laplacian(K, S, S_in, S_out)
    L *= 4 / sigma**4
    return L


def exp_laplacian(x_repr, x, sigma: float = 1, K=None):
    r"""
    Computation of the discrete Laplacian operator for the exponential kernel
    .. math::
        L[i,j] = \sum_{k} \nabla_{x} k(x_k, y_i)^\top \nabla_x k(x_i, y_j)

    Parameters
    ----------
    x_repr: ndarray of size (p, d)
        ::math:`x_repr = y` representer to discretize `L^2`
    x: ndarray of size (n, d)
        Data to estimate the distribution on `L^2(X)`
    sigma
        Bandwidth parameter for the kernel
    K: ndarray of size (p, n) (optional, default is None)
        Pre-computation of the gaussian kernel `k(x_repr, x)`

    Returns
    -------
    L: ndarray of size (p, p)
    """
    if K is None:
        K = exp_kernel(x_repr, x, sigma=sigma)
    S_in = scalar_product(x_repr, x_repr)
    S = scalar_product(x_repr, x)
    S_out = np.sum(x**2, axis=1)
    L = _exp_laplacian(K, S, S_in, S_out)
    L *= 1 / sigma**2
    return L


@numba.jit("f8[:, :](f8[:, :], f8[:, :], f8[:, :], f8[:])")
def _rbf_laplacian(K, S, S_in, S_out):
    p, n = K.shape
    out = np.zeros((p, p), dtype=np.float64)
    for i in range(p):
        for j in range(p):
            for k in range(n):
                tmp = S_out[k] - S[i, k] - S[j, k] + S_in[i, j]
                out[i, j] += K[i, k] * K[j, k] * tmp
    return out


@numba.jit("f8[:, :](f8[:, :], f8[:, :], f8[:, :], f8[:])")
def _exp_laplacian(K, S, S_in, S_out):
    p, n = K.shape
    out = np.zeros((p, p), dtype=np.float64)
    for i in range(p):
        for j in range(p):
            for k in range(n):
                norm = S_out[k] - 2 * S[i, k] + S_in[i, i]
                norm *= S_out[k] - 2 * S[j, k] + S_in[j, j]
                tmp = S_out[k] - S[i, k] - S[j, k] + S_in[i, j]
                if norm <= 0:
                    norm = 1
                    tmp = 0
                out[i, j] += K[i, k] * K[j, k] * tmp / np.sqrt(norm)
    return out



"""
Synthetic datasets

@ Vivien Cabannes, 2023
"""
import numpy as np


# Dataset
def halfmoons(n, noise_level: float = .1):
    """
    Return dataset in two dimensions populating two half-moons

    Parameters
    ----------
    n: int
        Number of sample
    noise_level
        Noise in the dataset, e.g. a value of .3 means overlap between clusters

    Return
    ------
    out: ndarray of size (n, 2)
        Dataset
    """
    theta = np.random.rand(n)
    theta *= 2 * np.pi
    out = np.empty((n, 2), dtype=float)
    out[:, 0] = np.cos(theta)
    out[:, 1] = np.sin(theta)
    out[out[:, 0] > 0, 1] += 1
    out += noise_level * np.random.randn(n, 2)
    return out


def concentric_circle(n, noise_level: float = .1, nb_circles: int = 4):
    """
    Return dataset in two dimensions populating concentric circles

    Parameters
    ----------
    n: int
        Number of sample
    noise_level
        Noise in the dataset, e.g. a value of .3 means overlap between clusters
    nb_circles
        Number of concentric circles in the dataset

    Return
    ------
    out: ndarray of size (n, 2)
        Dataset
    """
    theta = np.random.rand(n)
    theta *= 2 * np.pi
    out = np.empty((n, 2), dtype=float)
    out[:, 0] = np.cos(theta)
    out[:, 1] = np.sin(theta)
    out *= np.random.choice(nb_circles, n)[:, np.newaxis] + 1
    out += noise_level * np.random.randn(n, 2)
    return out



import numpy as np


def meshgrid_3d(n):
    theta = np.linspace(0, 2 * np.pi, n)
    phi = np.linspace(0, np.pi, n)
    X = np.outer(np.cos(theta), np.sin(phi))
    Y = np.outer(np.sin(theta), np.sin(phi))
    Z = np.outer(np.ones(n), np.cos(phi))
    return X, Y, Z



"""
Linear algebra utilities

@ Vivien Cabannes, 2023
"""
import numpy as np


def scalar_product(x1, x2):
    r"""
    Computation of
    .. math::
        out[i,j] = \sum_{k} x1[i, k] * x2[j,k]

    Parameters
    ----------
    x1: ndarray of size (n, d)
    x2: ndarray of size (n, d)

    Returns
    -------
    out: ndarray of size (n, n)
    """
    return x1 @ x2.T


def distance_square(x1, x2):
    r"""
    Computation of
    .. math::
        out[i,j] = \| x1[i, :] - x2[j, :] \|^2

    Parameters
    ----------
    x1: ndarray of size (n, d)
    x2: ndarray of size (n, d)

    Returns
    -------
    out: ndarray of size (n, n)
    """
    out = scalar_product(x1, x2)
    out *= -2
    out += np.sum(x1**2, axis=1)[:, np.newaxis]
    out += np.sum(x2**2, axis=1)
    out[out < 0] = 0
    return out


def distance(x1, x2):
    r"""
    Computation of
    .. math::
        out[i,j] = \| x1[i, :] - x2[j, :] \|

    Parameters
    ----------
    x1: ndarray of size (n, d)
    x2: ndarray of size (n, d)

    Returns
    -------
    out: ndarray of size (n, n)
    """
    out = distance_square(x1, x2)
    np.sqrt(out, out=out)
    return out
